"""
Web component
"""

import logging
from datetime import datetime

import yaml
from flask import Blueprint, Response, redirect, render_template
from flask_wtf import FlaskForm
from wtforms import DateTimeLocalField, IntegerField, StringField
from wtforms.validators import DataRequired, Length

from slinky import Slinky

slinky_webapp = Blueprint('webapp', __name__, template_folder='templates')

with open('config.yaml', encoding='utf-8-sig') as conffile:
    cfg = yaml.safe_load(conffile)


class ShortURLForm(FlaskForm):  # type: ignore[misc]
    """
    Web form definition
    """

    url = StringField(
        'URL',
        validators=[DataRequired(), Length(1, 2048)],
        render_kw={
            'size': 64,
            'maxlength': 2048,
            'placeholder': 'https://www.example.com',
        },
    )
    fixed_views = IntegerField(
        'Fixed number of views',
        validators=[DataRequired()],
        render_kw={
            'size': 3,
            'value': -1,
        },
    )
    length = IntegerField(
        'Shortcode length',
        validators=[DataRequired()],
        render_kw={
            'size': 3,
            'value': 4,
        },
    )
    expiry = DateTimeLocalField(
        'Expiry',
        format='%Y-%m-%dT%H:%M',
        render_kw={
            'size': 8,
            'maxlength': 10,
        },
    )


@slinky_webapp.route('/<path:path>')
def try_path_as_shortcode(path: str) -> Response:
    """
    Try the initial path as a shortcode, redirect if found

    Returns:
        Response: redirect if found, otherwise 404
    """
    should_redirect = True
    slinky = Slinky(cfg['db'])
    shortcode = slinky.get(path)
    if shortcode.url:
        if shortcode.fixed_views == 0:
            logging.warning('Shortcode out of views')
            should_redirect = False
        elif shortcode.fixed_views > 0:
            slinky.remove_view(shortcode.id)
        if datetime.fromisoformat(shortcode.expiry) < datetime.now():
            logging.warning('Shortcode expired')
            should_redirect = False

        if should_redirect:
            return redirect(shortcode.url, 302)

    return Response('Not found', 404)


@slinky_webapp.route('/_/add', methods=['GET', 'POST'])
def add() -> Response:
    """
    Create and add a new shorturl

    Returns:
        str: HTTP response
    """
    shortcode = ''
    url = ''

    form = ShortURLForm(meta={'csrf': False})

    if form.is_submitted():
        url = form.url.data.strip()
        length = form.length.data
        fixed_views = form.fixed_views.data
        expiry = form.expiry.data or datetime.max

    if url:
        slinky = Slinky(cfg['db'])
        for attempts in range(50):
            try:
                shortcode = slinky.add(url, length, fixed_views, expiry)
                break
            except ValueError:
                logging.warning(
                    'Shortcode already exists.  Retrying (%d/50).', attempts
                )
        else:
            return Response('Could not create a unique shortcode', 500)

    return render_template('add.html', form=form, shortcode=shortcode)


@slinky_webapp.route('/_/list', methods=['GET', 'POST'])
def lister() -> str:
    """
    Create and add a new shorturl

    Returns:
        str: shortcode for the URL
    """
    slinky = Slinky(cfg['db'])
    return render_template('list.html', shortcodes=slinky.get_all())
